/*
	File Name:	modules.c
*/

#include <StandardFile.h>
#include <FragLoad.h>
#include <MixedMode.h>
#include <ConditionalMacros.h>
#include "globs.h"
#include "prefs.h"
#include "modules.h"

/********************************************************************

	ExecuteSharedLib
	
	This code is only executed from the native application. It loads
	a shared library, looks up an exported symbol (function) name,
	creates a RoutineDescriptor to represent that function, and then
	executes the function. When complete, it closes the connection to
	the shared library.
	
********************************************************************/

#ifdef __powerc
OSErr ExecuteSharedLib(Str255 errName)
{
	OSErr			err = noErr, err2 = noErr;
	ConnectionID	connID = kNoConnectionID;
	Ptr				mainAddr = nil;
	
	//
	// attempt to locate and prepare the shared library
	//
	
	err = GetSharedLibrary("\pShared Library", kPowerPCArch, kLoadNewCopy,
						   &connID, &mainAddr, errName);
	if (err == noErr) {
		Ptr			symAddr = nil;
		SymClass	symClass;
		
		//
		// look up the symbol (in this case, a function) that we are interested in
		//
		
		err = FindSymbol(connID, "\pBeepThreeTimes", &symAddr, &symClass);
		if (err == noErr) {
			
			//
			// create a routine descriptor for the function, call it, 
			// then dispose the routine descriptor.
			//
			
			UniversalProcPtr upp = NewRoutineDescriptor((ProcPtr)symAddr, kCStackBased, GetCurrentISA());
			if (upp) {
				SetCursor(*GetCursor(watchCursor));
				CallUniversalProc(upp, kCStackBased);
				InitCursor();
				DisposeRoutineDescriptor(upp);
			} else err = memFullErr;
		}
		
		//
		// look up another symbol, just for fun.
		//
		
		if (err == noErr) {
			err = FindSymbol(connID, "\pShowLibAlert", &symAddr, &symClass);
			if (err == noErr) {
				
				//
				// create a routine descriptor for the function, call it, 
				// then dispose the routine descriptor.
				//
				
				UniversalProcPtr upp = NewRoutineDescriptor((ProcPtr)symAddr, kCStackBased, GetCurrentISA());
				if (upp) {
					CallUniversalProc(upp, kCStackBased);
					DisposeRoutineDescriptor(upp);
				} else err = memFullErr;
			}
		}
		
		//
		// close the connection
		//
		
		err2 = CloseConnection(&connID);
		if (err == noErr) err = err2;
	}
	
	return err;
}
#endif

/********************************************************************

	ExecuteResource
	
	This function is called on both platforms in order to execute a
	private resource.
	
********************************************************************/

OSErr ExecuteResource(void)
{
	FSSpec				codeSpec;
	Boolean				good;
	OSErr				err = noErr;
	short				saveResFile = CurResFile();
	short				fRefNum;
	
	// choose a valid file based on our preferences
	good = ChooseCodeContainingFile(&codeSpec);
	if (!good) return noErr;
	
	fRefNum = FSpOpenResFile(&codeSpec, fsRdPerm);
	if (fRefNum != -1) {
		UseResFile(fRefNum);
		
		switch(gCodeCallPreference) {
			case item68kRadio:
				err = Call68k();
				break;
			case itemPowerPCRadio:
				err = CallPPC();
				break;
			case itemNoModeSwitchRadio:
				err = CallFat();
				break;
			default:
				DebugStr("\pgCodeCallPreference is invalid!"); // should NEVER happen
				break;
		}
		
		CloseResFile(fRefNum);
		UseResFile(saveResFile);
	}
	
	return err;
}

/********************************************************************

	Call68k
	
	This function can be called from 68k code or PowerPC code.
	We use our handy macros in order to create the proper RoutineDescriptor
	when needed on the PowerMac. We have a special macro that returns
	a RoutineDescriptor to 68k code. We can then call the code with
	our CallModuleEntryProc function.
	
********************************************************************/

OSErr Call68k(void)
{
	long				moduleResult;
	ModuleEntryProcUPP	mepUPP;
	Handle 				hCode = Get1Resource(kModule68kResType, kModuleResID);
	OSErr				err = noErr;
	
	if (hCode) {
		HLockHi(hCode);
		
		mepUPP = NewModuleEntryProc68k(*hCode);
		moduleResult = CallModuleEntryProc(mepUPP, 0L);
		DisposeRoutineDescriptor(mepUPP);
		
		HUnlock(hCode);
		ReleaseResource(hCode);
	} else err = ResError();
	
	return err;
}

/********************************************************************

	CallPPC
	
	This function can be called from 68k code or PowerPC code.
	Since our PowerPC code resource already has a RoutineDescriptor
	tacked onto itself (by the linker) we can simply dereference the
	resource handle and call it.

********************************************************************/

OSErr CallPPC(void)
{
	long		moduleResult;
	Handle 		hCode = Get1Resource(kModulePPCResType, kModuleResID);
	OSErr		err = noErr;
	
	if (hCode) {
		HLockHi(hCode);
		
		moduleResult = CallModuleEntryProc((ModuleEntryProcUPP)*hCode, 0L);
		
		HUnlock(hCode);
		ReleaseResource(hCode);
	} else err = ResError();
	
	return err;
}

/********************************************************************

	CallFat
	
	This function can be called from 68k code or PowerPC code.
	Since our fat code resource already has a RoutineDescriptor
	tacked onto itself (by the linker) we can simply dereference the
	resource handle and call it.
	
********************************************************************/

OSErr CallFat(void)
{
	long				moduleResult;
	Handle 				hCode = Get1Resource(kModuleFatResType, kModuleResID);
	OSErr				err = noErr;
	
	if (hCode) {
		HLockHi(hCode);

		moduleResult = CallModuleEntryProc((ModuleEntryProcUPP)*hCode, 0L);
		
		HUnlock(hCode);
		ReleaseResource(hCode);
	} else err = ResError();
	
	return err;
}

/********************************************************************

	ChooseCodeContainingFile
	
	This function merely allows the user to choose which private
	resource they would like to execute.
	
********************************************************************/

Boolean ChooseCodeContainingFile(FSSpecPtr fsp)
{
	SFTypeList			typeList = {kModuleFileType};
	short				numTypes = 1;
	StandardFileReply	reply;
	FileFilterUPP		fileFilter;
	Boolean				result = false;
	
	fileFilter = NewFileFilterProc(MyFileFilter);
	if (fileFilter) {
		StandardGetFile(fileFilter, numTypes, typeList, &reply);
		result = reply.sfGood;
		DisposeRoutineDescriptor(fileFilter);
	}
	
	*fsp = reply.sfFile;
	return result;
}

/********************************************************************

	MyFileFilter
	
	This function filters files in our StandardGetFile dialog box.
	This way, we make sure the user selects a file that contains the
	proper type of code that we are looking for.
	
********************************************************************/

pascal Boolean MyFileFilter(CInfoPBPtr cpbp)
{
	FSSpec	fileSpec;
	Boolean	dontDisplayFile = true;
	OSErr	err;
	short	fRefNum;
	short	saveResFile = CurResFile();
	ResType	rType = '????';
	
	err = FSMakeFSSpec(cpbp->hFileInfo.ioVRefNum, cpbp->hFileInfo.ioFlParID,
					   cpbp->hFileInfo.ioNamePtr, &fileSpec);
	if (err != noErr) { SysBeep(0); return dontDisplayFile; }
	
	fRefNum = FSpOpenResFile(&fileSpec, fsRdPerm);
	if (fRefNum != -1) {
		UseResFile(fRefNum);
		
		switch(gCodeCallPreference) {
			case item68kRadio:
				rType = kModule68kResType;
				break;
			case itemPowerPCRadio:
				rType = kModulePPCResType;
				break;
			case itemNoModeSwitchRadio:
				rType = kModuleFatResType;
				break;
			default:
				DebugStr("\pgCodeCallPreference is invalid!"); // should NEVER happen
				break;
		}
		
		dontDisplayFile = (Count1Resources(rType) == 0);
		
		CloseResFile(fRefNum);
		UseResFile(saveResFile);
	}
	
	return dontDisplayFile;
}