// *****************************************************************************
//
//	WindoidDraw.c
//
// 
//	Copyright  1991-94 Infinity Systems.  All rights reserved.
// 
//	DESCRIPTION:
//		This file contains the code that the WDEF uses to get the locations
//		and sizes of parts, and to draw them.
// 
//	WRITTEN BY:
//		Troy Gaul
//		Infinity Systems
// *****************************************************************************

#include "WindoidDraw.h"	

#include <Memory.h>

#include "WindoidTypes.h"
#include "WindoidUtil.h"


// *****************************************************************************
//
//	Routines to get Rects for title bar parts													 
//
// *****************************************************************************

// 
//
//	GetTitleBarRect
//
// 
void 
GetTitleBarRect(WindowPeek window, Rect *titleBar) 
{
	*titleBar = (**window->strucRgn).rgnBBox;

	if (WindData.isHoriz) {
		// Titlebar on top.
		titleBar->bottom = titleBar->top + kTitleHeight;
		titleBar->right -= 1;		// Shadow compensation.
		
	} else {
		// Titlebar on left.
		titleBar->right = titleBar->left + kTitleHeight;
		titleBar->bottom -= 1; 		// Shadow compensation.
	}
}


// 
//
//	GetCloseBox
//
// 
void 
GetCloseBox(WindowPeek window, Rect *theRect) 
{
	GetTitleBarRect(window, theRect);

	if (WindData.isHoriz)
		InsetRect(theRect, kGadgetMargin, kGadgetInset);	// Titlebar on top.
	else
		InsetRect(theRect, kGadgetInset, kGadgetMargin);	// Titlebar on left.

	theRect->bottom = theRect->top  + kGadgetSize;
	theRect->right  = theRect->left + kGadgetSize;
}


// 
//
//	GetZoomBox
//
// 
void 
GetZoomBox(WindowPeek window, Rect *theRect) 
{
	GetTitleBarRect(window, theRect);

	if (WindData.isHoriz) {
		// Align zoom box with titlebar pattern.
		if (IsEven(theRect->right - theRect->left))
			OffsetRect(theRect, -1, 0);
		InsetRect(theRect, kGadgetMargin, kGadgetInset);	// Titlebar on top.
	
	} else {
		// Align zoom box with titlebar pattern.
		if (IsEven(theRect->bottom - theRect->top))
			OffsetRect(theRect, 0, -1);

		InsetRect(theRect, kGadgetInset, kGadgetMargin);	// Titlebar on left.
	}
	
	theRect->top  = theRect->bottom - kGadgetSize;
	theRect->left = theRect->right  - kGadgetSize;
}


// 
//
//	GetGrowBox
//
// 
void 
GetGrowBox(WindowPeek window, Rect *theRect) 
{
	GetGlobalContentRect(window, theRect);

	theRect->left = ++theRect->right  - kScrollBarPixels;
	theRect->top  = ++theRect->bottom - kScrollBarPixels;
}


// *****************************************************************************
//
//	Color Setup routines													 
//
// *****************************************************************************

// 
//
//	SetWFrameColor
//
// 
void
SetWFrameColor(WindowPeek window, Boolean isColor) 
{
	Boolean isActive = window->hilited || WindData.ignoreHilite;

	if (isColor) {
		if (isActive)
			WctbForeColor(window, wFrameColor);
		else
			AvgWctbForeColor(window, wHiliteColorLight, wHiliteColorDark,
							 wInactiveFramePct);
							 
	} else {
		ForeColor(blackColor);
	}
}


// 
//
//	SetWTitleColor
//
// 
void
SetWTitleColor(WindowPeek window, Boolean isColor) 
{
	Boolean isActive = window->hilited || WindData.ignoreHilite;

	if (isColor) {
		if (isActive)
			WctbForeColor(window, wTextColor);
		else		// Set the color for inactive titlebar text.
			AvgWctbForeColor(window, wHiliteColorLight, wHiliteColorDark,
							 wInactiveTextPct);
							 
	} else {
		ForeColor(blackColor);
	}
}


// 
//
//	SetWTitleBarColors
//
// 
//		Set the foreground and background for the drawing of the 
//		titlebar pattern.
// 
void
SetWTitleBarColors(WindowPeek window, Boolean isColor) 
{
	Boolean isActive = window->hilited || WindData.ignoreHilite;

	if (isColor) {
		if (isActive) {
			AvgWctbForeColor(window, wHiliteColorLight, wHiliteColorDark, 
							 wTitleBarDarkPct);
			AvgWctbBackColor(window, wHiliteColorLight, wHiliteColorDark, 
							 wTitleBarLightPct);
							 
		} else {
			WctbForeColor(window, wContentColor);
			WctbBackColor(window, wContentColor);
		}
		
	} else {
		ColorsNormal();
	}
}


// 
//
//	SetGadgetFrameEraseColors
//
// 
//		Set the foreground and background for the drawing of the 
//		titlebar pattern, in inverse so we can erase some of the
//		background by using normal drawing routines.
// 
void
SetGadgetFrameEraseColors(WindowPeek window, Boolean isColor) 
{
	if (isColor) {
		AvgWctbBackColor(window, wHiliteColorLight, wHiliteColorDark, 
						 wTitleBarDarkPct);
		AvgWctbForeColor(window, wHiliteColorLight, wHiliteColorDark, 
						 wTitleBarLightPct);
						 
	} else {
		ForeColor(whiteColor);
		BackColor(blackColor);
	}
}


// *****************************************************************************
//
//	Drawing routines													 
//
// *****************************************************************************

// 
//
//	DrawTitlebarTinges
//
// 
static void
DrawTitlebarTinges(WindowPeek window, Boolean isColor, const Rect *bounds) 
{
	Rect tempRect = *bounds;
	InsetRect(&tempRect, 1, 1);
	
	if (isColor) {
		AvgWctbForeColor(window, wTingeLight, wTingeDark, 
						 wTitleBarTingeDarkPct);
		FrameBottomRightShading(tempRect);

		WctbForeColor(window, wTingeLight);
		tempRect.right--;
		tempRect.bottom--;
		FrameTopLeftShading(tempRect);
		
	} else {
		ForeColor(whiteColor);
		BackColor(blackColor);
		FrameRect(&tempRect);
	}
}


// 
//
//	DrawCloseBox
//
// 
void 
DrawCloseBox(WindowPeek window, Boolean isColor, const Rect *theRect) 
{
	Rect tempRect;

	// Paint the area on the edges out with the background color.
	SetGadgetFrameEraseColors(window, isColor);
	
	tempRect = *theRect;
	InsetRect(&tempRect, -kTingeInset, -kTingeInset);
	FrameRect(&tempRect);

	if (isColor) {
		WctbForeColor(window, wTingeDark);
		BackColor(whiteColor);
		FrameTopLeftShading(*theRect);
		
		tempRect = *theRect;
		tempRect.top++;
		tempRect.left++;
		WctbForeColor(window, wTingeLight);
		FrameRect(&tempRect);

		InsetRect(&tempRect, 1, 1);
		WctbForeColor(window, wTingeDark);
		FrameBottomRightShading(tempRect);

		tempRect.right--;
		tempRect.bottom--;
		AvgWctbForeColor(window, wTitleBarLight, wTitleBarDark, 
						 wCloseBoxColor);
		PaintRect(&tempRect);
		
	} else {
		ColorsNormal();
		FrameBox(theRect);
	}
}


// 
//
//	DrawZoomBox
//
// 
void 
DrawZoomBox(WindowPeek window, Boolean isColor, const Rect *theRect) 
{
	Rect tempRect;

	DrawCloseBox(window, isColor, theRect);
	tempRect = *theRect;

	tempRect.bottom -= kGadgetSize / 2;
	tempRect.right -= kGadgetSize / 2;
		// This should handle gadgets of various sizes elegantly.

	if (isColor) {
		WctbForeColor(window, wTingeDark);
		tempRect.left += 2;		// Inset past the tinge.
		tempRect.top += 2;
		FrameBottomRightShading(tempRect);
		
	} else {
		FrameRect(&tempRect);
	}
}


// 
//
//	DrawXedBox
//
// 
//		Draw close or zoom box with an X in it (or inverted in B&W).
// 
void 
DrawXedBox(WindowPeek window, Boolean isColor, const Rect *theRect) 
{
	if (isColor) {
		AvgWctbForeColor(window, wTingeLight, wTingeDark, wXedBoxPct);
		PaintRect(theRect);
		
		WctbForeColor(window, wTitleBarDark);
		FrameRect(theRect);
		
		// Draw the X.
		MoveTo(theRect->left,      theRect->top       );
		LineTo(theRect->right - 1, theRect->bottom - 1);
		MoveTo(theRect->right - 1, theRect->top       );
		LineTo(theRect->left,      theRect->bottom - 1);

	} else {
		PaintRect(theRect);
	}
}


// 
//
//	DrawGrow3DBox
//
// 
static void 
DrawGrow3DBox(WindowPeek window, Rect *theRect, Boolean isLight) 
{
	Rect tempRect = *theRect;
	
	WctbForeColor(window, wTingeDark);
	FrameRect(theRect);
	
	// Add the top light outer border on the top-left edge.
	tempRect.left++;
	tempRect.top++;
	WctbForeColor(window, wTingeLight);
	FrameTopLeftShading(tempRect);

	// Finally, fill in the center.
	AvgWctbForeColor(window, wTitleBarLight, wTitleBarDark, 
					 isLight ? wGrowBoxColorLt : wGrowBoxColorDk);

	InsetRect(&tempRect, 1, 1);
	PaintRect(&tempRect);
}


// 
//
//	DrawGrowBox
//
// 
void 
DrawGrowBox(WindowPeek window, Boolean isColor) 
{
	Boolean isActive = window->hilited || WindData.ignoreHilite;
	Rect theRect;
	GetGrowBox(window, &theRect);

	if (!isActive) { 
		if (isColor) {
			WctbForeColor(window, wFrameColor);
			WctbBackColor(window, wContentColor);
		}
		FrameBox(&theRect);
		
	} else {
		Rect smallRect; 
		Rect largeRect;
	
		// Add the size box chevrons.
		SetRect(&smallRect, theRect.left + 3, theRect.top + 3,
						    theRect.left + 10, theRect.top + 10);
		SetRect(&largeRect, smallRect.left + 2, smallRect.top + 2, 
						    theRect.right - 2, theRect.bottom - 2);

		if (isColor) {
			WctbForeColor(window, wFrameColor);
			AvgWctbBackColor(window, wHiliteColorLight, wHiliteColorDark, 
							 wGrowBoxBackground);
			FrameBox(&theRect);

			// Draw the dark border parts for the bottom rectangle.
			OffsetRect(&largeRect, -1, -1);
			DrawGrow3DBox(window, &largeRect, false);

			// Draw the dark border parts for the top rectangle.
			smallRect.right--;
			smallRect.bottom--;
			DrawGrow3DBox(window, &smallRect, true);
			
		} else {
			ColorsNormal();
			FrameBox(&theRect);
			
			FrameRect(&largeRect);
			FrameBox(&smallRect);
		}
	}
}


// 
//
//	ToggleCloseBox
//
// 
void
ToggleCloseBox(WindowPeek window, Boolean isColor)
{
	Rect tempRect;
	GetCloseBox(window, &tempRect);
	
	if (WindData.closeToggle)
		DrawCloseBox(window, isColor, &tempRect);
	else
		DrawXedBox(window, isColor, &tempRect);
}


// 
//
//	ToggleZoomBox
//
// 
void
ToggleZoomBox(WindowPeek window, Boolean isColor)
{
	Rect tempRect;
	GetZoomBox(window, &tempRect);
	
	if (WindData.zoomToggle)
		DrawZoomBox(window, isColor, &tempRect);
	else
		DrawXedBox(window, isColor, &tempRect);
}


// 
//
//	GetTitlebarPat
//
// 
//		Choose correct pattern, depending on position of window in global
//		coordinates. (Concept of new (2.3) version taken from _Macintosh 
//		Programming Secrets_, Second Edition, by Scott Knaster and Keith 
//		Rollin, page 423.)
// 
static void
GetTitlebarPat(Boolean isActive, Point *corner, Pattern *titlePat) 
{
	long seed = isActive ? 0x00550055 : 0x00000000;

	if (IsOdd(corner->h))
		seed <<= 1;
	if (IsOdd(corner->v))
		seed <<= 8;

	*((long*) titlePat + 1) = *((long*) titlePat) = seed;
}


// 
//
//	SubtractGadgetRect
//
// 
static void
SubtractGadgetRect(RgnHandle theRgn, const Rect *theRect) 
{
	Rect subRect = *theRect;
	RgnHandle subRgn = NewRgn();
	
	InsetRect(&subRect, -kTingeInset, -kTingeInset);
		// To give the correct visual appearance.
	
	RectRgn(subRgn, &subRect);
	DiffRgn(theRgn, subRgn, theRgn);
	
	DisposeRgn(subRgn);
}


// 
//
//	DrawTitleString
//
// 
// 		When this routine is called, the background color will already be set 
// 		to the color of the background of the titlebar.
// 
static void
DrawTitleString(WindowPeek window, Boolean isColor, const Rect *titleRect, 
				Rect *stringRect) 
{
	Boolean isActive = window->hilited || WindData.ignoreHilite;
	short maxWidth;
	short titleWidth;
	short inset;
	short strAreaLeft;
	Rect titleStrBounds;
	RGBColor saveFore;
	
	if (window->titleHandle != nil && (*window->titleHandle)[0] != 0
		&& WindData.isHoriz) 
	{
		maxWidth = titleRect->right - titleRect->left - 2 * kGadgetMargin;
		strAreaLeft = titleRect->left + kGadgetMargin;
		if (window->goAwayFlag || window->spareFlag) {
			maxWidth -= 2 * (kGadgetSize + kGadgetMargin);
			strAreaLeft += kGadgetSize + kGadgetMargin;
		}
		
		if (maxWidth > 0) {
			char saveTitleHandleState;
				
			// Set up fonts, colors for text drawing.
			TextFont(kTitleFont);
			TextSize(kTitleSize);
			TextFace(kTitleStyle);
			TextMode(srcOr);
			
			saveTitleHandleState = HGetState((Handle) window->titleHandle);
			HLock((Handle) window->titleHandle);
				// StringWidth and DrawString may move memory.
			
			if (isColor)
				GetForeColor(&saveFore);
			SetWTitleColor(window, isColor);
			
			// Calculate the width of the title string.
			titleWidth = StringWidth(*window->titleHandle) + 2 * kTitleMargin;
			
			// Limit its size to maxWidth.
			titleWidth = (titleWidth > maxWidth) ? maxWidth : titleWidth;
			
			// Determine where to position it.
			inset = (short) (maxWidth - titleWidth) / 2;

			// Make the title appear centered.
			if (IsEven(titleWidth))				// We need an odd width or  
				titleWidth--;					// the overlap is wrong.

			inset -= IsEven(inset);
				// This is done so that the title doesnt 
				// shift as the windows width changes.
			
			// Set up the Rect to enclose the title within the titlebar.
			SetRect(&titleStrBounds, strAreaLeft + inset, 
									 titleRect->top + 1, 
									 strAreaLeft + inset + titleWidth, 
									 titleRect->bottom - 1); 
			
			// Inset the bounds so as not to erase part of the tinges.
			if (isActive)
				InsetRect(&titleStrBounds, 0, kTingeInset);
			
			// Return the title string areas boundry.
			*stringRect = titleStrBounds;
			
			// Make sure this area is cleared to the titlebar background color.
			EraseRect(&titleStrBounds);
			
			// Inset past the empty sides.
			InsetRect(&titleStrBounds, kTitleMargin, 0);

			// Outset it so decenders may overwrite the bottom tinge.
			if (isActive)
				InsetRect(&titleStrBounds, 0, -kTingeInset);

			// Clip the drawing to the strings area.
			{
				RgnHandle saveClip = NewRgn();
				RgnHandle clipRgn  = NewRgn();
		
				GetClip(saveClip);
				
				// Get the region the title string should go into.
				RectRgn(clipRgn, &titleStrBounds);
				
				// Make sure we dont clobber other windows.
				SectRgn(saveClip, clipRgn, clipRgn);
				if (!EmptyRgn(clipRgn)) {
					SetClip(clipRgn);
		
					// Draw the title.
					MoveTo(titleStrBounds.left, titleStrBounds.bottom - kTitleVDelta);
					DrawString(*window->titleHandle);
			
					// Clean up.
					SetClip(saveClip);
				}
				
				DisposeRgn(saveClip);
				DisposeRgn(clipRgn);
			}
			
			if (isColor)
				RGBForeColor(&saveFore);
			
			HSetState((Handle) window->titleHandle, saveTitleHandleState);

			TextFont(systemFont);
			TextSize(0);
			TextFace(0);
		}
	}
}


// 
//
//	DrawAndSubtractTitle
//
// 
//		Draws the title in the titlebar (by calling DrawTitleString) and 
//		subtracts it from the titlebar region (by making the rect returned
//		into a region and using DiffRgn).
// 
static void 
DrawAndSubtractTitle(WindowPeek window, const Rect *titleRect, Boolean isColor,
			 		 RgnHandle titleRgn) 
{
	Rect strRect;
	strRect.top = strRect.left = strRect.bottom = strRect.right = 0;
	
	// Draw the titlebar string (if any).
	DrawTitleString(window, isColor, titleRect, &strRect);
		// Since the area affected by the title string is returned in
		// strRect, we can use its left and right to draw the pattern.
	
	if (!EmptyRect(&strRect)) {
		RgnHandle tempRgn = NewRgn();
		
		RectRgn(tempRgn, &strRect);
		DiffRgn(titleRgn, tempRgn, titleRgn);
		
		DisposeRgn(tempRgn);
	}
}


// 
//
//	DrawTitleBar
//
// 
//		This routine actually draws the pattern into the titlebar. Note: it
//		takes a Rect as a parameter (not by address) because it goes ahead and
//		modifies it. I figured this was no worse than needing to copy it into
//		a local variable, so I went ahead and did it this way.
// 
void 
DrawTitleBar(WindowPeek window, Boolean isColor) 
{
	Boolean isHoriz = WindData.isHoriz;
	Boolean isActive = window->hilited || WindData.ignoreHilite;
	RgnHandle titleRgn = NewRgn();
	
	Rect titleRect;
	GetTitleBarRect(window, &titleRect);
	
	// Draw the frame.
	SetWFrameColor(window, isColor);
	FrameRect(&titleRect);
	
	// Set up the titleRgn region to be the whole titlebar, 
	// parts will then be punched out of it.
	{
		Rect insetTitleRect = titleRect;
		InsetRect(&insetTitleRect, 1 + (isActive ? kTingeInset : 0), 
								   1 + (isActive ? kTingeInset : 0));
			// Make room for the tinge, if any (which was already drawn).
		RectRgn(titleRgn, &insetTitleRect);
	}
		
	// Draw the tinges and gadgets.
	if (isActive) {
		Rect tempRect;

		// Draw the tinges.
		DrawTitlebarTinges(window, isColor, &titleRect);
	
		// Draw and subtract the close box.
		if (window->goAwayFlag) {
			GetCloseBox(window, &tempRect);
			DrawCloseBox(window, isColor, &tempRect);
			SubtractGadgetRect(titleRgn, &tempRect);
		}
	
		// Draw and subtract the zoom box.
		if (window->spareFlag) {
			GetZoomBox(window, &tempRect);
			DrawZoomBox(window, isColor, &tempRect);
			SubtractGadgetRect(titleRgn, &tempRect);
		}
	}

	// Get the colors to draw the rest.
	SetWTitleBarColors(window, isColor);

	// Draw the title.
	DrawAndSubtractTitle(window, &titleRect, isColor, titleRgn);

	// Draw the pattern.
	{
		Pattern pat;
		GetTitlebarPat(isActive, (Point*) &titleRect.top, &pat);
		FillRgn(titleRgn, &pat);
	}

	DisposeRgn(titleRgn);
}


// 
//
//	DrawWindowFrame
//
// 
void
DrawWindowFrame(WindowPeek window, Boolean isColor)
{
	Rect tempRect = (**window->strucRgn).rgnBBox;
	
	// Draw content frame and shadow.
	tempRect.bottom--;
	tempRect.right--;
	
	SetWFrameColor(window, isColor);
	FrameRect(&tempRect);
	
	// Draw Shadow.
	if (isColor)
		WctbForeColor(window, wFrameColor);		// This got messed up in 2.6.
	OffsetRect(&tempRect, 1, 1);
	FrameBottomRightShading(tempRect);
}


// 
