/*
	File Name:	PlayFromDisk XCMD.c
*/

/*************************************************************************

	PlayFromDisk XCMD

	by Joe Zobkiw
	
	Plays asynchronous sound from a disk-based AIFF file in HyperCard.
	
	Form: PlayFromDisk [fullPathName]
	
*************************************************************************/

#include <Sound.h>
#include <SoundInput.h>
#include <Script.h>
#include <HyperXCmd.h>
#include <string.h>
#include "A4Stuff.h"
#include "SetupA4.h"

/*************************************************************************

	#defines

*************************************************************************/
   
#define kMinParamCount	0
#define kMaxParamCount	1
#define kErrorFlag		(short)-1

#define kDefaultBufferSize	(1024*100L)	// try to allocate 100K buffer
#define kBufferDecrement 	(1024*5L)	// decrement buffer size by 5K
#define kSmallestBuffer		(1024*20L)	// but not below 20K

#define kGlobalIsPlayingFlag	"\pZobkiwIsPlaying"
#define kWindowName				"\pZobkiw"
#define kIdleTime				15	/* every x ticks */

/*************************************************************************

	function prototypes

*************************************************************************/
   
pascal 	void main(XCmdPtr xp);
pascal 	void MyCompletionRoutine(SndChannelPtr chan);
void 	HandleWindowMessage(XCmdPtr xp);
void	HandleXCMDMessage(XCmdPtr xp);
Handle	ConcatErrorStr(XCmdPtr xp, char *ch, OSErr err);
void	SetError(XCmdPtr xp, char *ch, OSErr err);
Handle	CopyStrToHand(char *ch);
long	HandleToNum(XCmdPtr xp, Handle h);
void	HandleToPStr(Str255 str, Handle h);
char*	ToCStr(char *ch);
char*	ToPStr(char *ch);

/*************************************************************************

	globals

*************************************************************************/

Boolean					gCloseFile;
short					gFileRefNum;
long					gSoundChannel;
FilePlayCompletionUPP	gFPCupp;

/*************************************************************************

	main

*************************************************************************/

pascal void main(XCmdPtr xp)
{
#ifndef powerc
	long	oldA4 = SetCurrentA4();
	RememberA4();
#endif
	
	if (xp->paramCount == -1) {
		HandleWindowMessage(xp);
	} else {
		HandleXCMDMessage(xp);
	}
	
#ifndef powerc
	SetA4(oldA4);
#endif
}

/*************************************************************************

	MyCompletionRoutine

	This tells us that the sound is done playing so we can close the
	audio file the next chance we get.
	
*************************************************************************/
   
pascal void MyCompletionRoutine(SndChannelPtr chan)
{
#ifndef powerc
	long oldA4 = SetUpA4();
#endif

	gCloseFile = true;
	
#ifndef powerc
	RestoreA4(oldA4);
#endif
}

/*************************************************************************

	HandleWindowMessage

	This function is called when our window gets a message. It will be called
	repeatedly with idle events while our sound is playing, since we created 
	an invisible window during that time. During these calls we simply check
	to see if the gCloseFile flag is set to true, which means our completion
	routine has been called which means the sound is done playing. When this
	occurs, we tell HyperCard to close the window.
	
	This in turn calls this function again with an xCloseEvt message. This
	is where we take care of disposing the sound channel, and reseting
	some of the flags that we used to tell HyperCard that we had interrupt
	code, we wanted idle time, and we were playing a sound.
	
*************************************************************************/

void HandleWindowMessage(XCmdPtr xp)
{
	XWEventInfoPtr	xw = (XWEventInfoPtr)(xp->params[0]);
	WindowPtr		w = xw->eventWindow;
	OSErr			err = noErr;
	
	if (xw->event.what == xOpenEvt) {
		;	// ignore this message
	} else if (xw->event.what == xCloseEvt) {
		// close the sound file
		if (gFileRefNum != -1) {
			FSClose(gFileRefNum);
			gFileRefNum = -1;
		}
		// dispose the sound channel
		if (gSoundChannel != nil) {
			err = SndDisposeChannel((SndChannelPtr)gSoundChannel, true);
			if (err != noErr)
				SetError(xp, "Error: Disposing sound channel. ", err);
			gSoundChannel = 0L;
		}

		// reset some flags and clean up globals
		SetXWIdleTime(xp, w, 0);
		XWHasInterruptCode(xp, w, false);
		EndXSound(xp);
		SetGlobal(xp, kGlobalIsPlayingFlag, CopyStrToHand("false"));
		xp->passFlag = true;
		
		if (gFPCupp) {
			DisposeRoutineDescriptor(gFPCupp);
			gFPCupp = nil;
		}
		
	} else if ((gCloseFile == true) && (gSoundChannel != nil)) {
		// the sound is done playing, tell the window to close
		gCloseFile = false;
		CloseXWindow(xp, w);
	}
}

/*************************************************************************

	HandleXCMDMessage
	
	This function basically exists to begin the playing of a sound.

*************************************************************************/

void HandleXCMDMessage(XCmdPtr xp)
{
	Str255				fileName;
	Str32				tempStr;
	StandardFileReply	reply;
	FSSpec				fileSpec;
	SFTypeList			typeList;
	long				bufferSize = kDefaultBufferSize;
	OSErr				err = noErr;
	short				refNum = -1;
	WindowPtr			w = nil;
	Rect				boundsRect = {0,0,33,33};
	SndChannelPtr		chan = nil;
	
	// tell HyperCard that we have interrupt code in this XCMD
	XWHasInterruptCode(xp, w, true);

	// check parameter count
	if ((xp->paramCount != kMinParamCount) && (xp->paramCount != kMaxParamCount)) {
		SetError(xp, "Error: Form = PlayFromDisk [fullPathName]. ", 0);
		goto fail;
	}
	
	// check global variable to see if we are already playing,
	// if so, stop the current sound, then do nothing.
	{
		Handle 	hGlob;
		Str32	tempStr;
		hGlob = GetGlobal(xp, kGlobalIsPlayingFlag);
		HandleToPStr(tempStr, hGlob);
		if (StrToBool(xp, tempStr)) {
			MyCompletionRoutine((SndChannelPtr)gSoundChannel);
			return;
		}
	}
	
	// initialize globals
	gCloseFile = false;
	gFileRefNum = -1;
	gSoundChannel = 0L;
	gFPCupp = NewFilePlayCompletionProc(MyCompletionRoutine);
	
	if (xp->paramCount == 0) {
		// first ask the user to select the file if no parameter
		typeList[0] = 'AIFF';
		typeList[1] = 'AIFC';
		StandardGetFile(nil, 2, typeList, &reply);
		if (!reply.sfGood) {
			SetError(xp, "cancel", 0);
			goto fail;
		}
		
		fileSpec = reply.sfFile;

		// attempt to update the screen
		ZeroToPas(xp, "Go to this card", tempStr);
		SendHCMessage(xp, tempStr);
	} else {
		// if the scripter passed in a file name, use it as a full path
		HandleToPStr(fileName, xp->params[0]);
		err = FSMakeFSSpec(0, 0, fileName, &fileSpec);
		if (err != noErr) {
			SetError(xp, "Error: Couldnt locate file. ", err);
			goto fail;
		}
	}
	
	// attempt to create a window with a unique name
	w = NewXWindow(xp, &boundsRect, kWindowName, false, documentProc, false, false);
	if (w == nil) {
		SetError(xp, "Error: Couldnt create window. ", 0);
		goto fail;
	}
	
	// open the file to play
	err = FSpOpenDF(&fileSpec, fsRdPerm, &refNum);
	if ((err != noErr) || (refNum == -1)) {
		SetError(xp, "Error: Couldnt open sound file. ", err);
		goto fail;
	}
	
	SetCursor(*GetCursor(watchCursor));
	
	// tell HyperCard that our window wants idle time every 1 tick
	SetXWIdleTime(xp, w, kIdleTime);
	
	// tell HyperCard that we are about to being playing a sound 
	// (ie: using a sound channel
	BeginXSound(xp, nil);

	// allocate our sound channel
	err = SndNewChannel(&chan, sampledSynth, 0, nil);
	if (err != noErr) {
		SetError(xp, "Error: Couldnt allocate sound channel. ", err);
		goto fail;
	}
		
	// remember refnum & snd chan
	gFileRefNum = refNum;
	gSoundChannel = (long)chan;

tryAgain:
	// play the sound file
	err = SndStartFilePlay(chan, refNum, 0, bufferSize, nil, nil, gFPCupp, true);
	if ((err == notEnoughBufferSpace) && (bufferSize > kSmallestBuffer)) {
		bufferSize -= kBufferDecrement;
		goto tryAgain;
	}
	
	if (err != noErr) {
		SetError(xp, "Error: Couldnt play the sound file. ", err);	
		goto fail;
	}
	
	// set a unique global variable to tell that we are now playing
	SetGlobal(xp, kGlobalIsPlayingFlag, CopyStrToHand("true"));
	
	goto exit;
fail:
	// clean up if fail
	gCloseFile = false;
	gFileRefNum = -1;
	gSoundChannel = 0L;
	if (gFPCupp) {
		DisposeRoutineDescriptor(gFPCupp);
		gFPCupp = nil;
	}
	SetXWIdleTime(xp, w, 0);
	XWHasInterruptCode(xp, w, false);
	EndXSound(xp);
	if (w)
		CloseXWindow(xp, w);
exit:
	InitCursor();
	return;
}

/*************************************************************************

	

*************************************************************************/

Handle	ConcatErrorStr(XCmdPtr xp, char *ch, OSErr err)
{
	Str255	str1;
	Str32	str2;
	
	if (err == noErr) {
		return((Handle)CopyStrToHand(ch));
	} else {
		strcpy((char*)str1, ch);
		NumToString((long)err, (StringPtr)&str2);
		ToCStr((char*)str2);
		strcat((char*)str1, (char*)str2);
		return((Handle)CopyStrToHand((char*)str1));
	}
}

/*************************************************************************

	

*************************************************************************/

void	SetError(XCmdPtr xp, char *ch, OSErr err)
{
	xp->returnValue = (Handle)ConcatErrorStr(xp, ch, err);
}

/*************************************************************************

	

*************************************************************************/

Handle	CopyStrToHand(char *ch)
{
	Handle	h;
	
	h = NewHandleClear((long)strlen(ch) + 1);
	if (h)
		strcpy((char*)*h, ch);
	return h;
}

/*************************************************************************

	

*************************************************************************/

long	HandleToNum(XCmdPtr xp, Handle h)
{
	char	str[32];
	long	num;
	
	strcpy(str, *h);
	num = StrToLong(xp, (StringPtr)ToPStr(str));
	return num;
}

/*************************************************************************

	

*************************************************************************/

void	HandleToPStr(Str255 str, Handle h)
{
	strcpy((char*)str, *h);
	ToPStr((char*)str);
}

/*************************************************************************

	

*************************************************************************/

char*	ToCStr(char *ch)
{
	unsigned char len, i;
	
	len = ch[0];
	for (i=0;i<len;++i)
		ch[i] = ch[i+1];
	ch[len] = 0;
	return ch;
}

/*************************************************************************

	

*************************************************************************/

char*	ToPStr(char *ch)
{
	unsigned char len, i;
	
	for (i=0,len=0;ch[i]!=0;++i)
		++len;
	while (i--)
		ch[i+1] = ch[i];
	ch[0] = len;
	return ch;
}